from discoPy.rest.base_request.base_request import BaseRequestAPI

class StageData(BaseRequestAPI):
    '''Contains a collection of stage related methods.'''

    def __init__(self, token: str, url: str=None):
        super().__init__(token, url)

    def create_stage_instance(self, channel_id, topic: str, privacy_level: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#create-stage-instance'''
        payload = {
            'channel_id': channel_id,
            'topic': topic
        }
        if privacy_level != None:
            payload['privacy_level'] = privacy_level
        return self._request('POST', params=payload, uri=f'/stage-instances')

    def get_stage_instance(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#get-stage-instance'''
        return self._request('GET', uri=f'/stage-instances/{channel_id}')

    def modify_stage_instance(self, topic: str=None, privacy_level=None) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#modify-stage-instance'''
        payload = {}
        if topic != None:
            payload['topic'] = topic
        if privacy_level != None:
            payload['privacy_level'] = privacy_level
        return self._request('PATCH', params=payload, uri=f'/stage-instances/{channel_id}')

    def delete_stage_instance(self) -> dict:
        '''https://discord.com/developers/docs/resources/stage-instance#delete-stage-instance'''
        return self._request('DELETE', uri=f'/stage-instances/{channel_id}')


from discoPy.rest.user.user import UserData
from discoPy.rest.channel.channel import ChannelData
from discoPy.rest.guild.guild import GuildData
from discoPy.rest.stage.stage import StageData
from discoPy.rest.webhook.webhook import WebhookData
from discoPy.rest.application.application import ApplicationData

class User(UserData):
    pass
class Guild(GuildData):
    pass
class Channel(ChannelData):
    pass
class Stage(StageData):
    pass
class Webhook(WebhookData):
    pass
class Application(ApplicationData):
    pass




import asyncio
import websockets as ws
import sys, traceback
import json
import time 
import logging 
import numpy as np

import requests

from discoPy.core.base_client import BaseClient

class WSClient(BaseClient):
    '''
        Connect to the Discord websocket-feed.

        @param: token (str): required 
        @param: url (str):  required
    '''
    _heartbeat_interval = 41250 / 1000 # in seconds
    _last_ping = None
    _sequence = None

    all_intents = [
        'GUILDS', 'GUILD_MEMBERS', 'GUILD_BANS', 'GUILD_EMOJIS_AND_STICKERS', 'GUILD_INTEGRATIONS', 'GUILD_WEBHOOKS',
        'GUILD_INVITES', 'GUILD_VOICE_STATES', 'GUILD_PRESENCES', 'GUILD_MESSAGES', 'GUILD_MESSAGE_REACTIONS',
        'GUILD_MESSAGE_TYPING', 'DIRECT_MESSAGES','DIRECT_MESSAGE_REACTIONS', 'GUILD_SCHEDULED_EVENTS', 'GUILD_SCHEDULED_EVENTS'
    ]

    _url = 'wss://gateway.discord.gg/?v=9&encording=json'

    def __init__(self, token: str, intents: list, api_version: int=None):

        super().__init__(token=token, api_version=api_version)

        self._log = logging.getLogger(__name__)

        self._token = token
        self._intents = np.sum([self._get_intents(intent) for intent in intents])

        if api_version != None:
            self._url = f'wss://gateway.discord.gg/?v={api_version}&encording=json'
            
    def run(self) -> None:
        self._connect()

    async def _main(self) -> None:
        try:
            keepwaiting = True
            async with ws.connect(self._url) as socket:
                self.ws = socket
                await self._authenticate() 

                try:
                    while keepwaiting:
                        if not self._last_ping or int(time.time()) - self._last_ping > self._get_ws_pingtimeout():
                            await self._send_ping()
                        try:
                            evt = await asyncio.wait_for(self.ws.recv(), timeout=self._get_ws_pingtimeout())
                        except asyncio.TimeoutError:
                            self._log.info(f'No message in {self._get_ws_pingtimeout()} seconds')
                            await self._send_ping()
                        except asyncio.CancelledError:
                            self._log.warn(f'Cancelled error')
                            await self.ws.ping()
                        else:
                            try:
                                evt_obj = json.loads(evt)
                                if evt_obj['op'] != 11 and 's' in evt_obj and evt_obj['s']: # setting the _sequence
                                    self._sequence = evt_obj['s']
                                    if evt_obj['op'] == 10:  
                                        self._heartbeat_interval = int(evt_obj['d']['heartbeat_interval']) / 1000
                            except ValueError:
                                pass
                            else:
                                await self.trigger('on_event', evt_obj)
                    
                except ws.ConnectionClosed:
                    keep_waiting = False
                    self._log.warn(f'Websocket connection closed!  {e} {traceback.format_exc()}')
                    await self._reconnect()
                except Exception as e:
                    self._log.error(f'WS exception: {e} {traceback.format_exc()}')
                    keep_waiting = False
                    await self._reconnect()

        except Exception as e:
            self._log.error(f'Exception!: {e} {traceback.format_exc()}')

    def _connect(self) -> None:
        self._log.info('Connecting...')
        self._conn = asyncio.ensure_future(self._main())

    async def _reconnect(self) -> None:
        self._log.info('Reconnecting Websocket...')
        await self._close_connection()
        self._connect()
    
    async def _close_connection(self) -> None:
        await self.ws.close()

    async def send_message(self, message: dict) -> None:
        await self.ws.send(json.dumps(message))

    async def _send_ping(self):
        self._log.info(f'Sending ping now, _sequence: {self._sequence}')
        await self.send_message({
            'op': 1,
            'd': self._sequence
        })
        self._last_ping = int(time.time())
    
    async def _authenticate(self) -> None:
        await self.send_message(self._get_payload(kind='init'))

    def _get_ws_pingtimeout(self) -> None:
        return self._heartbeat_interval

    def _get_payload(self, kind: str) -> dict:
        '''https://discord.com/developers/docs/topics/gateway#identifying'''
        op = None
        if kind == 'reconnect':
            op = 7
        elif kind == 'resume':
            op = 6
        elif kind == 'init':
            op = 2
        else:
            raise ValueError
        return {
            'op': op,
            'd': {
                'token': self._token,
                'intents': f'{self._intents}',
                'properties': {
                    '$os': sys.platform,
                    '$browser': 'disco',
                    '$device': 'pc'
                },
                'compress': True
            }
        }
    @classmethod
    def get_intents_list(cls) -> list:
        return cls.all_intents

    def _get_intents(self, intents: str) -> int:
        # https://discord.com/developers/docs/topics/gateway#gateway-intents
        if intents == 'GUILDS':
            '''
                - GUILD_CREATE
                - GUILD_UPDATE
                - GUILD_DELETE
                - GUILD_ROLE_CREATE
                - GUILD_ROLE_UPDATE
                - GUILD_ROLE_DELETE
                - CHANNEL_CREATE
                - CHANNEL_UPDATE
                - CHANNEL_DELETE
                - CHANNEL_PINS_UPDATE
                - THREAD_CREATE
                - THREAD_UPDATE
                - THREAD_DELETE
                - THREAD_LIST_SYNC
                - THREAD_MEMBER_UPDATE
                - THREAD_MEMBERS_UPDATE
                - STAGE_INSTANCE_CREATE
                - STAGE_INSTANCE_UPDATE
                - STAGE_INSTANCE_DELETE
            '''
            return (1 << 0)
        elif intents == 'GUILD_MEMBERS':
            '''
                - GUILD_MEMBER_ADD
                - GUILD_MEMBER_UPDATE
                - GUILD_MEMBER_REMOVE
                - THREAD_MEMBERS_UPDATE *
            '''
            return (1 << 1)
        elif intents == 'GUILD_BANS':
            '''
                - GUILD_BAN_ADD
                - GUILD_BAN_REMOVE
            '''
            return (1 << 2)
        elif intents == 'GUILD_EMOJIS_AND_STICKERS':
            '''
                - GUILD_EMOJIS_UPDATE
                - GUILD_STICKERS_UPDATE
            '''
            return (1 << 3)
        elif intents == 'GUILD_INTEGRATIONS': 
            '''
                - GUILD_INTEGRATIONS_UPDATE
                - INTEGRATION_CREATE
                - INTEGRATION_UPDATE
                - INTEGRATION_DELETE
            '''
            return (1 << 4)
        elif intents == 'GUILD_WEBHOOKS':
            '''
                - WEBHOOKS_UPDATE
            '''
            return (1 << 5)
        elif intents == 'GUILD_INVITES':
            '''
                - INVITE_CREATE
                - INVITE_DELETE
            '''
            return (1 << 6)
        elif intents == 'GUILD_VOICE_STATES': 
            '''
                - VOICE_STATE_UPDATE
            '''
            return (1 << 7)
        elif intents == 'GUILD_PRESENCES': 
            '''
                - PRESENCE_UPDATE
            '''
            return (1 << 8)
        elif intents == 'GUILD_MESSAGES':
            '''
                - MESSAGE_CREATE
                - MESSAGE_UPDATE
                - MESSAGE_DELETE
                - MESSAGE_DELETE_BULK
            '''
            return  (1 << 9)
        elif intents == 'GUILD_MESSAGE_REACTIONS':
            '''
                - MESSAGE_REACTION_ADD
                - MESSAGE_REACTION_REMOVE
                - MESSAGE_REACTION_REMOVE_ALL
                - MESSAGE_REACTION_REMOVE_EMOJI
            '''
            return  (1 << 10)
        elif intents == 'GUILD_MESSAGE_TYPING': 
            '''
                - TYPING_START
            '''
            return (1 << 11)
        elif intents == 'DIRECT_MESSAGES':
            '''
                - MESSAGE_CREATE
                - MESSAGE_UPDATE
                - MESSAGE_DELETE
                - CHANNEL_PINS_UPDATE
            '''
            return (1 << 12)
        elif intents == 'DIRECT_MESSAGE_REACTIONS': 
            '''
                - MESSAGE_REACTION_ADD
                - MESSAGE_REACTION_REMOVE
                - MESSAGE_REACTION_REMOVE_ALL
                - MESSAGE_REACTION_REMOVE_EMOJI
            '''
            return (1 << 13)
        elif intents == 'DIRECT_MESSAGE_TYPING': 
            '''
                - TYPING_START
            '''
            return (1 << 14)
        elif intents == 'GUILD_SCHEDULED_EVENTS': 
            '''
                - GUILD_SCHEDULED_EVENT_CREATE
                - GUILD_SCHEDULED_EVENT_UPDATE
                - GUILD_SCHEDULED_EVENT_DELETE
                - GUILD_SCHEDULED_EVENT_USER_ADD
                - GUILD_SCHEDULED_EVENT_USER_REMOVE
            '''
            return (1 << 16)
        else:
            raise Exception('Uknown Intent!')



import asyncio
import logging
# import re
import sys
# from typing import IO
import urllib.error
import urllib.parse
import json
import aiohttp
from aiohttp import ClientSession

logging.basicConfig(
    format='%(asctime)s %(levelname)s:%(name)s: %(message)s',
    level=logging.WARNING,
    datefmt='%H:%M:%S',
    stream=sys.stderr,
)
logger = logging.getLogger(__name__)
logging.getLogger('chardet.charsetprober').disabled = True

from discoPy.core.application.application import Application
from discoPy.core.guild.guild import Guild
from discoPy.core.channel.channel import Channel
from discoPy.core.stage.stage import Stage
from discoPy.core.user.user import User
from discoPy.core.webhook.webhook import Webhook

class BaseClient(Application, Guild, Channel, Stage, User, Webhook):

    BASE_URL = 'https://discord.com/api/v9'
    TIMEOUT = 10
    headers = { 
        'content-type': 'application/json',
        'user-agent': 'python-discord-client'
    }

    def __init__(self, token: str, api_version: int=None) -> None:
        self._token = token
        if api_version != None:
            self.BASE_URL = f'https://discord.com/api/v{api_version}'
        
        self.headers['Authorization'] =  f'Bot {self._token}'

        self._callbacks = { 
            'on_event': [self.on_event] 
        }

    async def _request(self, method: str, params: dict={}, uri: str='', headers: dict={}, files: dict=None, **kwargs) -> dict:
        data_json = ''
        if method in ['GET', 'DELETE']:
            if params:
                strl = []
                for key in sorted(params):
                    strl.append('{}={}'.format(key, params[key]))
                data_json += '&'.join(strl)
                uri += f'?{data_json}'
        else:
            if params:
                data_json = params 
        try:
            payload = json.dumps(data_json)
        except:
            payload = data_json

        url = f'{self.BASE_URL}{uri}'
        response = None
        try:
            async with ClientSession(headers=self.headers) as session:
                if method == 'GET':
                    response = await session.request(method=method, url=url, headers=headers, timeout=self.TIMEOUT, **kwargs)
                elif method == 'DELETE':
                    response = await session.request(method=method, url=url, headers=headers, timeout=self.TIMEOUT, **kwargs)
                elif method == 'POST':
                    response = await session.request(method=method, url=url, data=payload, headers=headers,  timeout=self.TIMEOUT, **kwargs)
                elif method == 'PATCH':
                    response = await session.request(method=method, url=url, data=payload, headers=headers, timeout=self.TIMEOUT, **kwargs)
                elif method == 'PUT':
                    response = await session.request(method=method, url=url, data=payload, headers=headers, timeout=self.TIMEOUT, **kwargs)
        except (
            aiohttp.ClientError,
            aiohttp.http_exceptions.HttpProcessingError,
        ) as e:
            logger.error(
                'aiohttp exception for %s [%s]: %s',
                url,
                getattr(e, 'status', None),
                getattr(e, 'message', None),
            )
        except Exception as e:
            logger.exception('Non-aiohttp exception occured:  %s', getattr(e, '__dict__', {}))
        else:
            if response != None:
                return await self._check_response(response)
            else:
                print('!')

    async def _check_response(self, response) -> dict:
        if response.status == 200:
            try:
                 data = await response.json()
            except ValueError:
                raise Exception(response.text())
            else:
                return data
        else:
            raise Exception(f'{response.status}: {await response.text()}')


    async def _send_file_attachment(self, method: str, uri: str, file_names: [str], payload: dict={}) -> dict:
        raise NotImplementedError('Sorry, this is not working the asynchron way. Use the Rest endpoint instead! (discopy.rest.channel)')
        headers = { 'content-disposition': 'form-data; name="payload_json"', 'content-type': 'multipart/form-data' }
        payload = { 'payload_json': json.dumps(payload) }
        
        
        prepared_files = {}
        # for i, filename in enumerate(file_names):
        #     prepared_files[f'files[{i}]'] = (filename, open(filename, 'rb'), {
        #         'Content-Disposition': f'form-data; name="files[{i}]"; filename="{filename}"'
        #         }
        #     )

        # for key in prepared_files:
        #     payload[key] = prepared_files[key]

        print(payload)

        with aiohttp.MultipartWriter('mixed') as mp:
            y = mp.append_json(payload)
            x = mp.append(open('test.png','rb').read(), {
                'content-type': 'image/png',
                'Content-Disposition': 'form-data; name="files[0]"; filename="test.png"'
            })
        
            # payload2 = aiohttp.payload.get_payload(open('test.png', 'rb'))#, headers=headers)
            # payload2.set_content_disposition(
            #     'form-data',
            #     name='files[0]',
            #     filename='test.png'
            # )
            # payload2._headers['Content-Disposition'] = 'form-data; name="files[0]"; filename="test.png"'
            # Add part to multipart
            # part = mp.append(payload2)

        return await self._request(method, params=mp,  headers=headers, uri=uri)


    async def on(self, event_name: str, callback):
        if event_name not in self._callbacks:
            self._callbacks[event_name] = [callback]
        else:
            self._callbacks[event_name].append(callback)

    async def trigger(self, event_name: str, event_object):
        if event_name in self._callbacks:
            for callback in self._callbacks[event_name]:
                await callback(event_object)

    async def on_event(self, event) -> None:
        pass




import requests
import json

class BaseRequestAPI(object):

    BASE_URL = 'https://discord.com/api/v9'
    TIMEOUT = 10

    def __init__(self, token: str, url: str=None):
        self._token = token
        if url != None:
            self.BASE_URL = url
        self._session = self._get_session()

    def _request(self, method: str, params={}, uri: str='', headers: dict={}, files:dict=None) -> dict:
        uri_path = uri
        data_json = ''
        if method in ['GET', 'DELETE']:
            if params:
                strl = []
                for key in sorted(params):
                    strl.append('{}={}'.format(key, params[key]))
                data_json += '&'.join(strl)
                uri += f'?{data_json}'
        else:
            if params:
                data_json = params 
        try:
            payload = json.dumps(data_json)
        except:
            payload = data_json

        url = f'{self.BASE_URL}{uri}'
        response = None

        if method == 'GET':
            response = self._session.get(url=url, headers=headers, timeout=self.TIMEOUT)
        elif method == 'DELETE':
            response = self._session.delete(url=url, headers=headers, timeout=self.TIMEOUT)
        elif method == 'POST':
            response = self._session.post(url=url, data=payload, headers=headers, files=files, timeout=self.TIMEOUT)
        elif method == 'PATCH':
            response = self._session.patch(url=url, data=payload, headers=headers, timeout=self.TIMEOUT)
        elif method == 'PUT':
            response = self._session.put(url=url, data=payload, headers=headers, timeout=self.TIMEOUT)
        if response != None:
            return self.check_response(response)

    def _get_session(self) -> requests.Session:
        session = requests.Session()
        session.headers.update({
            'authorization': f'Bot {self._token}',
            'content-type': 'application/json',
            'user-agent': 'python-discord-client'
        })

        return session

    def _send_file_attachment(self, method: str, uri: str, file_names: [str], payload: dict={}, headers: dict={}) -> dict:
        self._session.headers.update({ 'content-type': None })
        headers = { 'content-disposition': 'form-data; name="payload_json"' }
        payload = { 'payload_json': json.dumps(payload) }
        prepared_files = {}
        for i, filename in enumerate(file_names):
            file_type = filename.split('.')[-1]
            media_type=None
            if file_type in ['jpg', 'png', 'jpeg', 'gif']:
                media_type = 'image'
                if file_type == 'svg':
                    file_type = 'svg+xml'

            prepared_files[f'files[{i}]'] = (filename, open(filename, 'rb'), f'{media_type}/{file_type}', {
                'Content-Disposition': f'form-data; name="files[{i}]"; filename="{filename}"'
                }
            )

        for key in prepared_files:
            payload[key] = prepared_files[key]

        response = self._request(method, params=payload, files=prepared_files, headers=headers, uri=uri)
        self._session.headers.update({ 'content-type': 'application/json' })
        return response

    @staticmethod
    def check_response(response_data) -> dict:
        if response_data.status_code == 200:
            try:
                data = response_data.json()
                # print(response_data.request.prepare_headers)
            except ValueError:
                raise Exception(response_data.content)
            else:
                return data
        else:
            raise Exception(f'{response_data.status_code}: {response_data.text}')


from discoPy.rest.base_request.base_request import BaseRequestAPI


class GuildData(BaseRequestAPI):
    '''Contains a collection of guild/server related methods.'''
    
    data = None
    def __init__(self, token: str, guild_id=None, url: str=None):
        super().__init__(token, url)
        if guild_id:
            self.data = self.get_channel(guild_id=channel_id)
            # print(self.data)

    # -----======= G E N E R A L =======-----
    def create_guild(self, 
        name: str, 
        region: str=None, 
        icon=None, 
        verification_level: int=None, 
        default_message_notifications: int=None, 
        explicit_content_filter: int=None, 
        roles: list=None, 
        channels: list=None, 
        afk_channel_id: str=None, 
        afk_timeout: int=None, 
        system_channel_id: str=None, 
        system_channel_flags: int=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#create-guild'''
        payload = { 'name': name }
        if region:
            payload['region'] = region
        if icon:
            payload['icon'] = icon
        if verification_level:
            payload['verification_level'] = verification_level
        if default_message_notifications:
            payload['default_message_notifications'] = default_message_notifications
        if explicit_content_filter:
            payload['explicit_content_filter'] = explicit_content_filter
        if roles:
            payload['roles'] = roles
        if channels:
            payload['channels'] = channels
        if afk_channel_id:
            payload['afk_channel_id'] = afk_channel_id
        if afk_timeout:
            payload['afk_timeout'] = afk_timeout
        if system_channel_id:
            payload['system_channel_id'] = system_channel_id
        if system_channel_flags:
            payload['system_channel_flags'] = system_channel_flags
            
        return self._request('POST', params = payload, uri='/guilds')

    def get_guild(self, guild_id, with_counts: bool=False, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild'''
        params = { 'width_couunts': with_counts }
        return self._request('GET', params=params, uri=f'/guilds/{guild_id}')

    def get_guild_preview(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-preview'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/preview')

    def modify_guild(self,
        name: str,
        region: str=None,
        verification_level: int=None,
        default_message_notifications: int=None,
        explicit_content_filter: int=None,
        afk_channel_id: str=None,
        afk_timeout: int=None,
        icon=None, 
        owner_id=None,
        splash=None,
        discovery_splash=None,
        banner=None,
        system_channel_id=None,
        system_channel_flags: int=None,rules_channel_id=None,
        public_updates_channel_id=None,
        preferred_locale: str=None,
        features: list=None,
        description: str=None,premium_progress_bar_enabled: bool=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild'''
        payload = { 'name': name }
        if region != None:
            payload['region'] = region
        if verification_level != None:
            payload['verification_level'] = verification_level
        if default_message_notifications != None:
            payload['default_message_notifications'] = default_message_notifications
        if explicit_content_filter != None:
            payload['explicit_content_filter'] = explicit_content_filter
        if afk_channel_id != None:
            payload['afk_channel_id'] = afk_channel_id
        if afk_timeout != None:
            payload['afk_timeout'] = afk_timeout
        if icon != None:
            payload['icon'] = icon
        if owner_id != None:
            payload['owner_id'] = owner_id
        if splash != None:
            payload['splash'] = splash
        if discovery_splash != None:
            payload['discovery_splash'] = discovery_splash
        if banner != None:
            payload['banner'] = banner
        if system_channel_id != None:
            payload['system_channel_id'] = system_channel_id
        if system_channel_flags != None:
            payload['system_channel_flags'] = system_channel_flags
        if public_updates_channel_id != None:
            payload['public_updates_channel_id'] = public_updates_channel_id
        if preferred_locale != None:
            payload['preferred_locale'] = preferred_locale
        if features != None:
            payload['features'] = features
        if description != None:
            payload['description'] = description

        return self._request(method='POST', params=payload, uri=f'/guilds/{guild_id}')

    def delete_guild(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#delete-guild'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}')

    def get_guild_channels(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-channels'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/channels')

    def create_guild_channel(self, 
        guild_id,
        name: str,
        type: int,
        topic: str,
        bitrate: int,
        user_limit: int,
        rate_limit_per_user: int,
        position: int,
        permission_overwrites: list,
        parent_id,
        nsfw: bool
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#create-guild-channel'''
        payload = {
            'guild_id': guild_id,
            'name': name,
            'type': type,
            'topic': topic,
            'bitrate': bitrate,
            'user_limit': user_limit,
            'rate_limit_per_user': rate_limit_per_user,
            'position': position,
            'permission_overwrites': permission_overwrites,
            'parent_id': parent_id,
            'nsfw': nsfw
        }
        return self._request(method='POST', params=payload, uri=f'/guilds/{guild_id}/channels')

    def modify_channel_positions(self, 
        guild_id, id, position: int=None, lock_permissions: bool=None, parent_id=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild-channel-positions'''
        payload = { 'id': id }
        if position:
            payload['position'] = position
        if lock_permissions:
            payload['lock_permissions'] = lock_permissions
        if parent_id:
            payload['parent_id'] = parent_id
        return self._request(method='PATCH', params=payload, uri=f'/guilds/{guild_id}/channels')

    def list_active_threads(self, guild_id, threads: list, members: list, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#list-active-threads'''
        payload = { 'threads': threads, 'members': members }
        return self._request(method='GET', params=payload, uri=f'/guilds/{guild_id}/threads/active')

    def get_guild_member(self, guild_id, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-member'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/members/{user_id}')

    def list_guild_members(self, guild_id, limit: int, after, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#list-guild-members'''
        params = {}
        if limit != None: 
            params['limit'] = limit
        if after != None:
            params['after'] = after
        return self._request(method='GET', params=params, uri=f'/guilds/{guild_id}/members')

    def search_guild_members(self, guild_id, query: str, limit:int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#search-guild-members'''
        params = { 'query': query }
        if limit != None: 
            params['limit'] = limit
        return self._request(method='GET', params=params, uri=f'/guilds/{guild_id}/members/search')

    def add_guild_member(self, guild_id, user_id, access_token: str, nick: str, roles: list, mute: bool, deaf: bool, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#add-guild-member'''
        payload = {
            'access_token': access_token,
            'nick': nick,
            'roles': roles,
            'mute': mute,
            'deaf': deaf
        }
        return self._request(method='PUT', params=payload, uri=f'/guilds/{guild_id}/members/{user_id}')

    def modify_current_member(self, guild_id, nick: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-current-member'''
        payload = {}
        if nick != None:
            payload['nick'] = nick
        return self._request(method='PATCH', params=payload, uri=f'/guilds/{guild_id}/members/@me')

    def add_guild_member_role(self, guild_id, user_id, role_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#add-guild-member-role'''
        return self._request(method='PUT', uri=f'/guilds/{guild_id}/members/{user_id}/roles/{role_id}')

    def remove_guild_member_role(self, guild_id, user_id, role_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#remove-guild-member-role'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/members/{user_id}/roles/{role_id}')

    def remove_guild_member(self, guild_id, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#remove-guild-member'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/members/{user_id}')

    def get_guild_bans(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-bans'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/bans')

    def get_guild_ban(self, guild_id, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-ban'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/bans/{user_id}')

    def create_guild_ban(self, guild_id, user_id, delete_message_days: int=None, reason: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#create-guild-ban'''
        params = {}
        if delete_message_days != None:
            params['delete_message_days'] = delete_message_days
        if reason != None:
            params['reason'] = reason
        return self._request(method='PUT', params=params, uri=f'/guilds/{guild_id}/bans/{user_id}')

    def remove_guild_ban(self, guild_id, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#remove-guild-ban'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/bans/{user_id}')

    def get_guild_roles(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-roles'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/roles')

    def create_guild_role(self, 
        guild_id, name: str, permissions: str, 
        color: int, hoist: bool, icon, unicode_emoji: str, 
        mentionable: bool
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#create-guild-role'''
        payload = {
            'name': name,
            'permissions': permissions,
            'color': color,
            'hoist':hoist,
            'icon': icon,
            'unicode_emoji': unicode_emoji,
            'mentionable': mentionable
        }
        return self._request(method='POST', params=payload, uri=f'/guilds/{guild_id}/roles')

    def modify_guild_role_permissions(self, guild_id, id, position: int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild-role-positions'''
        payload = { 'id': id }
        if position != None:
            payload['position'] = position
        return self._request(method='PATCH', params=payload, uri=f'/guilds/{guild_id}/roles')

    def modify_guild_role(self, 
        guild_id, role_id, name: str=None, permission: str=None, 
        color: int=None, hoist: bool=None, icon=None, unicode_emoji: str=None, mentionable: bool=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild-role'''
        payload = {}
        if name != None:
            payload['name'] = name
        if permission != None:
            payload['permission'] = permission
        if color != None:
            payload['color'] = color
        if hoist != None:
            payload['hoist'] = hoist
        if icon != None:
            payload['icon'] = icon
        if unicode_emoji != None:
            payload['unicode_emoji'] = unicode_emoji
        if mentionable != None:
            payload['mentionable'] = mentionable
        return self._request(method='PATCH', params=payload, uri=f'/guilds/{guild_id}/roles/{role_id}')

    def delete_guild_role(self, guild_id, role_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#delete-guild-role'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/roles/{role_id}')

    def get_guild_prune_count(self, guild_id, days: int=7, include_roles: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-prune-count'''
        params = {}
        if days != None:
            params['days'] = days
        if include_roles != None:
            params['include_roles'] = include_roles
        return self._request(method='GET', params=params, uri=f'/guilds/{guild_id}/prune')

    def beginn_guild_prune(self, guild_id, days: int, compute_prune_count: bool, include_roles: str, reason: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#begin-guild-prune'''
        payload = {
            'days': days,
            'compute_prune_count': compute_prune_count,
            'include_roles': include_roles
        }
        if reason != None:
            payload['reason'] = reason
        return self._request(method='POST', params=payload, uri=f'/guilds/{guild_id}/prune')

    def get_guild_voice_regions(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-voice-regions'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/regions')

    def get_guild_invites(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-invites'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/invites')

    def get_guild_integrations(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-integrations'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/integrations')

    def delete_guild_integrations(self, guild_id, integration_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#delete-guild-integration'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/integrations/{integration_id}')

    def get_guild_widget_settings(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-widget-settings'''
        return self._request(method='DELETE', uri=f'/guilds/{guild_id}/widget')

    def modify_guild_widget(self, guild_id, settings: dict, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild-widget'''
        return self._request(method='PATCH', params=settings, uri=f'/guilds/{guild_id}/widget')

    def get_guild_widget(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-widget'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/widget.json')

    def get_guild_vanity_url(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-vanity-url'''
        return self._request(method='GET', uri=f'/guilds/{guild_id}/vanity-url')

    def get_guild_widget_image(self, guild_id, style: str, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-widget-image
        @param style can be one of : [shield, banner1, banner2, banner3, banner4]
        '''
        params = { 'style': style }
        return self._request(method='GET', params=params, uri=f'/guilds/{guild_id}/widget.png')

    def get_guild_welcome_screen(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#get-guild-welcome-screen'''
        return self._request(method='GET', params=params, uri=f'/guilds/{guild_id}/welcome-screen')

    def modify_guild_welcome_screen(self, guild_id, enabled: bool=None, welcome_channels: list=None, description: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-guild-welcome-screen'''
        params = {}
        if enabled != None:
            params['enabled'] = enabled
        if welcome_channels != None:
            params['welcome_channels'] = welcome_channels
        if description != None:
            params['description'] = description
        return self._request(method='PATCH', params=params, uri=f'/guilds/{guild_id}/welcome-screen')

    def modify_current_user_voice_state(self, guild_id, channel_id, suppress: bool=None, request_to_speak_timestamp=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-current-user-voice-state'''
        params = { 'channel_id': channel_id }
        if suppress != None:
            params['suppress'] = suppress
        if request_to_speak_timestamp != None:
            params['request_to_speak_timestamp'] = request_to_speak_timestamp
        return self._request(method='PATCH', params=params, uri=f'/guilds/{guild_id}/voice-states/@me')
    
    def modify_user_voice_state(self, guild_id, user_id, channel_id, suppress: bool=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild#modify-user-voice-state'''
        params = { 'channel_id': channel_id }
        if suppress != None:
            params['suppress'] = suppress
        return self._request(method='PATCH', params=params, uri=f'/guilds/{guild_id}/voice-states/{user_id}e')

    def get_guild_audit_log(self, guild_id, user_id, action_type:int, before=None, limit: int=50, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/audit-log#get-guild-audit-log'''
        params = {}
        if user_id != None:
            params['user_id'] = user_id
        if action_type != None:
            params['action_type'] = action_type
        if before != None:
            params['before'] = before
        if limit != None:
            params['limit'] = limit
        return self._request('GET', params=params, uri=f'/guilds/{guild_id}/audit-logs')

    # -----======= S H E D U L E D - E V E N T S =======-----
    def list_scheduled_events_for_guild(self, guild_id, with_user_count:bool=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#list-scheduled-events-for-guild'''
        params = {}
        if with_user_count != None:
            params['with_user_count'] = with_user_count
        return self._request('GET', params=params, uri=f'/guilds/{guild_id}/scheduled-events')

    def create_guild_scheduled_event(self, 
        guild_id,
        name: str,
        entity_type,
        privacy_level,
        scheduled_start_time,
        scheduled_end_time,
        channel_id=None,
        entity_metadata=None,
        description: str=None,
        image=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#create-guild-scheduled-event'''
        payload = {
            'name': name,
            'entity_type': entity_type,
            'privacy_level': privacy_level,
            'scheduled_start_time': scheduled_start_time,
            'scheduled_end_time': scheduled_end_time
        }
        if channel_id != None:
            payload['channel_id'] = channel_id
        if entity_metadata != None:
            payload['entity_metadata'] = entity_metadata
        if description != None:
            payload['description'] = description
        if image != None:
            payload['image'] = image
        return self._request('POST', params=payload, uri=f'/guilds/{guild_id}/scheduled-events')

    def get_guild_scheduled_event(self, guild_id, guild_scheduled_event_id, with_user_count: bool=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#get-guild-scheduled-event'''
        params = {}
        if with_user_count != None:
            params['with_user_count'] = with_user_count
        return self._request('GET', params=params, uri=f'/guilds/{guild_id}/scheduled-events/{guild_scheduled_event_id}')

    def modify_guild_scheduled_event(self, 
        guild_id, 
        guild_scheduled_event_id,
        channel_id=None,
        entity_metadata=None,
        name: str=None,
        privacy_level=None,
        scheduled_start_time=None,
        scheduled_end_time=None,
        description: str=None,
        entity_type=None,
        status=None,
        image=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#modify-guild-scheduled-event'''
        payload = {}
        if channel_id != None:
            payload['channel_id'] = channel_id
        if entity_metadata != None:
            payload['entity_metadata'] = entity_metadata
        if name != None:
            payload['name'] = name
        if privacy_level != None:
            payload['privacy_level'] = privacy_level
        if scheduled_start_time != None:
            payload['scheduled_start_time'] = scheduled_start_time
        if scheduled_end_time != None:
            payload['scheduled_end_time'] = scheduled_end_time
        if description != None:
            payload['description'] = description
        if entity_type != None:
            payload['entity_type'] = entity_type
        if status != None:
            payload['status'] = status
        if image != None:
            payload['image'] = image

        return self._request('PATCH', params=payload, uri=f'/guilds/{guild_id}/scheduled-events/{guild_scheduled_event_id}')

    def delete_guild_scheduled_evend(self, guild_id, guild_scheduled_event_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#delete-guild-scheduled-event'''
        return self._request('DELETE', uri=f'/guilds/{guild_id}/scheduled-events/{guild_scheduled_event_id}')

    def get_guild_scheduled_event_users(self, 
        guild_id, 
        guild_scheduled_event_id,
        limit: int=None,
        with_member: bool=None,
        before=None,
        after=None
    , **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-scheduled-event#get-guild-scheduled-event-users'''
        params = {}
        if limit != None:
            params['limit'] = limit
        if with_member != None:
            params['with_member'] = with_member
        if before != None:
            params['before'] = before
        if after != None:
            params['after'] = after
        return self._request('GET', params=params, uri=f'/guilds/{guild_id}/scheduled-events/{guild_scheduled_event_id}/users')

    # -----======= T E M P L A T E S =======-----
    def get_guild_template(self, template_code, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#get-guild-template'''
        return self._request('GET', uri=f'/guilds/templates/{template_code}')

    def create_guild_template(self, template_code, name: str, icon=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#create-guild-from-guild-template'''
        return self._request('POST', uri=f'/guilds/templates/{template_code}')

    def get_guild_templates(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#get-guild-templates'''
        return self._request('GET', uri=f'/guilds/{guild_id}/templates')

    def create_guild_template(self, guild_id, name: str, description: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#create-guild-template'''
        payload = { 'name': name}
        if description != None:
            payload['description'] = description
        return self._request('POST', params=payload, uri=f'/guilds/{guild_id}/templates')

    def sync_guild_template(self, guild_id, template_code, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#sync-guild-template'''
        return self._request('PUT',  uri=f'/guilds/{guild_id}/templates/{template_code}')

    def modify_guild_template(self, name: str=None, description: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#modify-guild-template'''
        payload = {}
        if name != None:
            payload['name'] = name
        if description != None:
            payload['description'] = description
        return self._request('PATCH', params=payload, uri=f'/guilds/{guild_id}/templates/{template_code}')

    def delete_guild_template(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/guild-template#delete-guild-template'''
        return self._request('DELETE', uri=f'/guilds/{guild_id}/templates/{template_code}')


    # -----======= E M O J I =======-----
    def list_guild_emojis(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/emoji#list-guild-emojis'''
        return self._request('GET', uri=f'/guilds/{guild_id}/emojis')
        
    def get_guild_emoji(self, guild_id, emoji_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/emoji#get-guild-emoji'''
        return self._request('GET', uri=f'/guilds/{guild_id}/emojis/{emoji_id}')

    def create_guild_emoji(self, guild_id, name: str, image, roles: list, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/emoji#create-guild-emoji'''
        payload = {
            'name': name,
            'image': image,
            'roles': roles
        }
        return self._request('POST', params=payload, uri=f'/guilds/{guild_id}/emojis')

    def modify_guild_emoji(self, guild_id, emoji_id, name: str=None, roles: list=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/emoji#modify-guild-emoji'''
        payload = {}
        if name != None:
            payload['name'] = name
        if roles != None:
            payload['roles'] = roles
        return self._request('PATCH', params=payload, uri=f'/guilds/{guild_id}/emojis/{emoji_id}')

    def delete_guild_emoji(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/emoji#delete-guild-emoji'''
        return self._request('DELETE', uri=f'/guilds/{guild_id}/emojis/{emoji_id}')


    # -----======= S T I C K E R  =======-----
    def get_sticker(self, sticker_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#get-sticker'''
        return self._request('GET', uri=f'/stickers/{sticker_id}')
        
    def list_nitro_sticker_packs(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#list-nitro-sticker-packs'''
        return self._request('GET', uri=f'/sticker-packs')

    def list_guild_stickers(self, guild_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#list-guild-stickers'''
        return self._request('GET', uri=f'/guilds/{guild_id}/stickers')

    def get_guild_sticker(self, guild_id, sticker_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#get-guild-sticker'''
        return self._request('GET', uri=f'/guilds/{guild_id}/stickers/{sticker_id}')

    def create_guild_sticker(self, guild_id, name: str, description: str, tags: str, file, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#create-guild-sticker'''
        payload = {
            'name': name,
            'description': description,
            'tags': tags,
            'file': file
        }
        return self._request('POST', params=payload, uri=f'/guilds/{guild_id}/stickers')

    def modify_guild_sticker(self, name: str=None, description: str=None, tags: str=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#modify-guild-sticker'''
        payload = {}
        if name != None:
            payload['name'] = name
        if description != None:
            payload['description'] = description
        if tags != None:
            payload['tags'] = tags
        return self._request('PATCH', params=payload, uri=f'/guilds/{guild_id}/stickers/{sticker_id}')

    def delete_guild_sticker(self, guild_id, sticker_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/sticker#delete-guild-sticker'''
        return self._request('DELETE', uri=f'/guilds/{guild_id}/stickers/{sticker_id}')

    def _get_guild_id(self, kwargs) -> str:
        try:
            guild_id = kwargs.get('guild_id', None)
            return guild_id if guild_id != None else self.data['id']
        except:
            raise ValueError('guild_id not defined!')


class Channel(object):
    '''Contains a collection of channel related methods.'''

    async def get_channel(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel'''
        return await self._request('GET', uri=f'/channels/{channel_id}')

    async def modify_channel(self, 
        name: str=None,     
        icon=None, 
        type: int=None,
        position: int=None,
        topic: str=None,
        nsfw: bool=None,
        rate_limit_per_user: int=None,
        bitrate: int=None,
        user_limit: int=None,
        permission_overwrites: list=None,
        parent_id=None,
        rtc_region: str=None,
        video_quality_mode: int=None,
        default_auto_archive_duration: int=None,
        archived: bool=None,
        auto_archive_duration: int=None,
        locked: bool=None,
        invitable: bool=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#modify-channel
            
            ----- P A R M E T E R S -----
            If Channel == Group DM
                @name
                @icon
            If Channel == Guild Channel
                @name
                @type
                @position
                @topic
                @nsfw
                @rate_limit_per_user
                @bitrate
                @user_limit
                @permission_overwrites
                @parent_id
                @rtc_region
                @video_quality_mode
                @default_auto_archive_duration
            If Chanel == Thread
                @name
                @archived
                @auto_archive_duration
                @locked
                @invitable
                @rate_limit_per_user
        '''
        payload = {}
        if name != None:
            payload['name'] = name
        if icon != None:
            payload['icon'] = icon
        if type != None:
            payload['type'] = type
        if topic != None:
            payload['topic'] = topic
        if nsfw != None:
            payload['nsfw'] = nsfw
        if rate_limit_per_user != None:
            payload['rate_limit_per_user'] = rate_limit_per_user
        if bitrate != None:
            payload['bitrate'] = bitrate
        if user_limit != None:
            payload['user_limit'] = user_limit
        if permission_overwrites != None:
            payload['permission_overwrites'] = permission_overwrites
        if parent_id != None:
            payload['parent_id'] = parent_id
        if rtc_region != None:
            payload['rtc_region'] = rtc_region
        if video_quality_mode != None:
            payload['video_quality_mode'] = video_quality_mode
        if default_auto_archive_duration != None:
            payload['default_auto_archive_duration'] = default_auto_archive_duration
        if archived != None:
            payload['archived'] = archived
        if auto_archive_duration != None:
            payload['auto_archive_duration'] = auto_archive_duration
        if locked != None:
            payload['locked'] = locked
        if invitable != None:
            payload['invitable'] = invitable
        return await  self._request('PATCH', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}')
        
    async def delete_close_channel(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#deleteclose-channel'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}')

    async def get_channel_messages(self,
        around=None,
        before=None,
        after=None,
        limit: int=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-messages'''
        params = {}
        if around != None:
            params['around'] = around
        if before != None:
            params['before'] = beforde
        if after != None:
            params['after'] = after
        if limit != None:
            params['limit'] = limit
        return await  self._request('GET', params=params, uri=f'/channels/{await self._get_channel_id(kwargs)}/messages')

    async def get_channel_message(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-message'''
        return await  self._request('GET',uri=f'/channels/{await self._get_channel_id(kwargs)}/messages')

    async def create_message(self,
        content: str=None,
        tts: bool=None,
        embeds: [dict]=None,
        allowed_mentions=None,
        message_reference=None,
        components: list=None,
        files: [str]=None,
        attachments: [dict]=None,
        sticker_ids: list=None,
        flags: int=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-message'''

        payload = {}
        if content != None:
            payload['content'] = content
        if tts != None:
            payload['tts'] = tts
        if allowed_mentions != None:
            payload['allowed_mentions'] = allowed_mentions
        if message_reference != None:
            payload['message_reference'] = message_reference
        if components != None:
            payload['components'] = components
        if sticker_ids != None:
            payload['sticker_ids'] = sticker_ids
        if flags != None:
            payload['flags'] = flags
        if embeds != None:
            payload['embeds'] = embeds
        if attachments != None:
            payload['attachments'] = attachments
        
        uri = f'/channels/{await self._get_channel_id(kwargs)}/messages'
        if files != None:
            return await  self._send_file_attachment(method='POST', uri=uri, file_names=files, payload=payload)
        else:
            return await  self._request(method='POST', uri=uri, params=payload)

    async def crosspost_message(self, message_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#crosspost-message'''
        response = self._request('POST', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/crosspost')

    async def create_reaction(self, message_id, emoji, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        response = self._request('PUT', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions/{emoji}/@me')

    async def delete_own_reaction(self,  message_id, emoji, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-own-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        response = self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions/{emoji}/@me')

    async def delete_user_reaction(self, message_id, user_id, emoji, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-user-reaction
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        response = self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions/{emoji}/{user_id}')

    async def get_reactions(self, message_id, emoji, after=None, limit: int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-reactions
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        response = self._request('GET', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions/{emoji}')

    async def delete_all_reactions(self, message_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-all-reactions'''
        response = self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions')
        
    async def delete_all_reactions_for_emoji(self, message_id, emoji, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-all-reactions-for-emoji
            "The emoji must be URL Encoded or the request will fail with 10014: Unknown Emoji. 
            To use custom emoji, you must encode it in the format name:id with the emoji name and emoji id."
        '''
        response = self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/reactions/{emoji}')

    async def edit_message(self,
        message_id, 
        content: str=None,
        embeds: [dict]=None,
        flags: int=None,
        allowed_mentions=None,
        components=list,
        files: [str]=None,
        payload_json: str=None,
        attachments: list=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#edit-message'''
        payload = {}
        if content != None:
            payload['content'] = content
        if flags != None:
            payload['flags'] = flags
        if allowed_mentions != None:
            payload['allowed_mentions'] = allowed_mentions
        if components != None:
            payload['components'] = components
        if payload_json != None:
            payload['payload_json'] = payload_json
        if attachments != None:
            payload['attachments'] = attachments
        if embeds != None:
            payload['embeds'] = embeds

        uri = f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}'
        if files != None:
            return await  self._send_file_attachment(method='POST', uri=uri, file_names=files, payload=payload)
        else:
            return await  self._request(method='POST', uri=uri, payload=payload)

    async def delete_message(self, message_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-message'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}')

    async def bulk_delete_message(self, messages: list, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#bulk-delete-messages'''
        paylaod = { 'messages': messages }
        return await  self._request('POST', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/bulk-delete')

    async def edit_channel_permissions(self,overwrite_id, allow=None, deny=None, type: int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#edit-channel-permissions'''
        payload = {}
        if allow != None:
            payload['allow'] = allow
        if deny != None:
            payload['deny'] = deny
        if type != None:
            payload['type'] = type
        return await  self._request('PUT', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/permissions/{overwrite_id}')

    async def get_channel_invites(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-channel-invites'''
        return await  self._request('GET', uri=f'/channels/{await self._get_channel_id(kwargs)}/invites')

    async def create_channel_invite(self,
        max_age: int=None,
        max_uses: int=None,
        temporary: bool=None,
        unique: bool=None,
        target_type: int=None,
        target_user_id=None,
        target_application_id=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#create-channel-invite'''
        payload = {}
        if max_age != None:
            payload['max_age'] = max_age
        if max_uses != None:
            payload['max_uses'] = max_uses
        if temporary != None:
            payload['temporary'] = temporary
        if unique != None:
            payload['unique'] = unique
        if target_type != None:
            payload['target_type'] = target_type
        if target_user_id != None:
            payload['target_user_id'] = target_user_id
        if target_application_id != None:
            payload['target_application_id'] = target_application_id
        return await  self._request('POST', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/invites')

    async def delete_channel_permission(self, overwrite_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#delete-channel-permission'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/permissions/{overwrite_id}')

    async def follow_news_channel(self, webhook_channel_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#follow-news-channel'''
        payload = { 'webhook_channel_id': webhook_channel_id }
        return await  self._request('POST', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/followers')

    async def trigger_typing_indicatoe(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#trigger-typing-indicator'''
        return await  self._request('POST', uri=f'/channels/{await self._get_channel_id(kwargs)}/typing')

    async def get_pinned_message(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-pinned-messages'''
        return await  self._request('GET', uri=f'/channels/{await self._get_channel_id(kwargs)}/pins')

    async def pin_message(self, message_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#pin-message'''
        return await  self._request('PUT', uri=f'/channels/{await self._get_channel_id(kwargs)}/pins/{message_id}')

    async def unpin_message(self, message_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#unpin-message'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/pins/{message_id}')

    async def group_dm_add_recipient(self, user_id, access_token, nick, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#group-dm-add-recipient'''
        payload = {
            'access_token': access_token,
            'nick':nick
        }
        return await  self._request('PUT', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/recipients/{user_id}')

    async def group_dm_remove_recipient(self, user_i, **kwargsd) -> dict:
        '''https://discord.com/developers/docs/resources/channel#group-dm-remove-recipient'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/recipients/{user_id}')

    async def start_thread_with_message(self, user_id, name: str, auto_archive_duration: int=None, rate_limit_per_user: int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#start-thread-with-message'''
        payload = { 'name': name }
        if auto_archive_duration != None:
            payload['auto_archive_duration'] = auto_archive_duration
        if rate_limit_per_user != None:
            payload['rate_limit_per_user'] = rate_limit_per_user
        return await  self._request('POST', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/messages/{message_id}/threads')

    async def start_thread_without_message(self, 
        auto_archive_duration: str=None,
        type: int=None,
        invitable: bool=None,
        rate_limit_per_user: int=None,
        **kwargs
    ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#start-thread-without-message'''
        payload = { 'name': name }
        if auto_archive_duration != None:
            payload['auto_archive_duration'] = auto_archive_duration
        if type != None:
            payload['type'] = type
        if invitable != None:
            payload['invitable'] = invitable
        if rate_limit_per_user != None:
            payload['rate_limit_per_user'] = rate_limit_per_user
        return await  self._request('POST', params=payload, uri=f'/channels/{await self._get_channel_id(kwargs)}/threads')

    async def join_thread(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#join-thread'''
        return await  self._request('PUT', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members/@me')

    async def add_thread_member(self, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#add-thread-member'''
        return await  self._request('PUT', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members/{user_id}')

    async def leave_thread(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#leave-thread'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members/@me')

    async def remove_thread_member(self, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#remove-thread-member'''
        return await  self._request('DELETE', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members/{user_id}')

    async def get_thread_member(self, user_id, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#get-thread-member'''
        return await  self._request('GET', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members/{user_id}')

    async def list_thread_member(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-thread-members'''
        return await  self._request('GET', uri=f'/channels/{await self._get_channel_id(kwargs)}/thread-members')

    async def list_active_threads(self,  threads: list, members: list, has_more: bool, **kwargs ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-active-threads'''
        params = {
            'threads': threads,
            'members': members,
            'has_more': has_more
        }
        return await  self._request('GET', params=params, uri=f'/channels/{await self._get_channel_id(kwargs)}/threads/active')

    async def list_public_archived_threads(self, before=None, limit:int=None, **kwargs ) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-public-archived-threads'''
        params = {}
        if before != None:
            params['before'] = before
        if limit != None:
            params['limit'] = limit
        return await  self._request('GET', params=params, uri=f'/channels/{await self._get_channel_id(kwargs)}/threads/archived/public')

    async def list_private_archived_threads(self, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-private-archived-threads'''
        params = {}
        if before != None:
            params['before'] = before
        if limit != None:
            params['limit'] = limit
        return await  self._request('GET', params=params, uri=f'/channels/{await self._get_channel_id(kwargs)}/threads/archived/private')

    async def list_joined_private_archived_threads(self, before=None, limit:int=None, **kwargs) -> dict:
        '''https://discord.com/developers/docs/resources/channel#list-joined-private-archived-threads'''
        params = {}
        if before != None:
            params['before'] = before
        if limit != None:
            params['limit'] = limit
        return await  self._request('GET', params=params, uri=f'/channels/{await self._get_channel_id(kwargs)}/users/@me/threads/archived/private')

    async def _get_channel_id(self, kwargs) -> str:
        try:
            channel_id = kwargs.get('channel_id', None)
            return  channel_id if channel_id != None else self.data['id']
        except:
            raise ValueError('channel_id not defined!')


from discoPy.rest.base_request.base_request import BaseRequestAPI

class UserData(BaseRequestAPI):
    '''Contains a collection of user related methods.'''
    
    def __init__(self, token: str, url: str=None):
        super().__init__(token, url)

    def get_current_user(self) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user'''
        return self._request(method='GET', uri=f'/users/@me')

    def get_user(self, user_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-user'''
        return self._request(method='GET', uri=f'/users/{user_id}')

    def modify_current_user(self, username: str=None, avatar=None) -> dict:
        '''https://discord.com/developers/docs/resources/user#modify-current-user'''
        payload = { }
        if username != None:
            payload['username'] = username
        if avatar != None:
            payload['avatar'] = avatar
        return self._request(method='PATCH', params=payload, uri=f'/users/@me')

    def get_current_user_guilds(self, before=None, after=None, limit: int=None) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user-guilds'''
        params = { }
        if before != None:
            params['before'] = before
        if after != None:
            params['after'] = after
        if limit != None:
            params['limit'] = limit
        return self._request(method='GET', params=params, uri=f'/users/@me/guilds')

    def get_current_user_guild_member(self, guild_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-current-user-guild-member'''
        return self._request(method='GET', uri=f'/users/@me/guilds/{guild_id}/member')

    def leave_guild(self, guild_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#leave-guild'''
        return self._request(method='DELETE', uri=f'/users/@me/guilds/{guild_id}')

    def create_dm(self, recipient_id) -> dict:
        '''https://discord.com/developers/docs/resources/user#create-dm'''
        payload = { 'recipient_id': recipient_id }
        return self._request(method='POST', params=payload, uri=f'/users/@me/channels')

    def create_group_dm(self, access_tokens: list, nicks: dict) -> dict:
        '''https://discord.com/developers/docs/resources/user#create-group-dm'''
        payload = {
            'access_tokens': access_tokens,
            'nicks': nicks
        }
        return self._request(method='POST', params=payload, uri=f'/users/@me/channels')

    def get_user_connections(self) -> dict:
        '''https://discord.com/developers/docs/resources/user#get-user-connections'''
        return self._request(method='GET', uri=f'/users/@me/connections')

    def list_voice_regions(self) -> dict:
        '''https://discord.com/developers/docs/resources/voice#list-voice-regions'''
        return self._request(method='GET', uri=f'/voice/regions')

    def get_invite(self, invite_code, with_counts: bool=None, with_expiration: bool=None, guild_scheduled_event_id=None) -> dict:
        '''https://discord.com/developers/docs/resources/invite#get-invite'''
        params = {}
        if with_counts != None:
            params['with_counts'] = with_counts
        if with_expiration != None:
            params['with_expiration'] = with_expiration
        if guild_scheduled_event_id != None:
            params['guild_scheduled_event_id'] = guild_scheduled_event_id
        return self._request(method='GET', params=params, uri=f'/invites/{invite_code}')

    def delete_invite(self, invite_code) -> dict:
        '''https://discord.com/developers/docs/resources/invite#delete-invite'''
        return self._request(method='DELETE', params=params, uri=f'/invites/{invite_code}')



    


from discoPy.rest.base_request.base_request import BaseRequestAPI

class ApplicationData(BaseRequestAPI):
    '''Contains a collection of application related methods.'''

    def __init__(self, token: str, url: str=None):
        super().__init__(token, url)

    def get_application_commands(self, application_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-global-application-commands'''
        return self._request('GET', uri=f'/applications/{application_id}/commands')

    def create_application_command(self,
        application_id,
        name: str,
        description: str,
        options: list=None,
        default_permission: bool=None,
        type=None
    ) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#create-global-application-command'''
        payload = {
            'name': name,
            'description': description
        }
        if options != None:
            payload['options'] = options
        if default_permission != None:
            payload['default_permission'] = default_permission
        if type != None:
            payload['type'] = type
        return self._request('POST', params=payload, uri=f'/applications/{application_id}/commands')

    def get_global_application_command(self, application_id, command_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-global-application-command'''
        return self._request('GET', uri=f'/applications/{application_id}/commands/{command_id}')

    def edit_global_application_command(self, 
        application_id,
        command_id,
        name: str=None,
        description: str=None,
        options: list=None,
        default_permission: bool=None
    ) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#edit-global-application-command'''
        payload = {}
        if name != None:
            payload['name'] = name
        if description != None:
            payload['description'] = description
        if options != None:
            payload['options'] = options
        if default_permission != None:
            payload['default_permission'] = default_permission
        return self._request('PATCH', params=payload, uri=f'/applications/{application_id}/commands/{command_id}')

    def delete_global_application_command(self, application_id, command_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#delete-global-application-command'''
        return self._request('DELETE', uri=f'/applications/{application_id}/commands/{command_id}')

    def bulk_overwrite_global_application_command(self, application_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#bulk-overwrite-global-application-commands'''
        return self._request('PUT', uri=f'/applications/{application_id}/commands')

    def get_guild_application_commands(self, application_id, guild_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-guild-application-commands'''
        return self._request('GET', uri=f'/applications/{application_id}/guilds/{guild_id}/commands')

    def create_guild_application_command(self, 
        application_id,
        guild_id,
        name: str,
        description: str,
        options: list=None,
        default_permission: bool=None,
        type=None
    ) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#create-guild-application-command'''
        payload = {
            'name': name,
            'description': description
        }
        if options != None:
            payload['options'] = options
        if default_permission != None:
            payload['default_permission'] = default_permission
        if type != None:
            payload['type'] = type
        return self._request('POST', params=payload,uri=f'/applications/{application_id}/guilds/{guild_id}/commands')

    def get_guild_application_command(self, application_id, guild_id, command_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-guild-application-command'''
        return self._request('GET', uri=f'/applications/{application_id}/guilds/{guild_id}/commands/{command_id}')

    def edit_guild_application_command(self,
        application_id, 
        guild_id,
        command_id,
        name: str=None,
        description: str=None,
        options: list=None,
        default_permission: bool=None
    ) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#edit-guild-application-command'''
        payload = {}
        if name != None:
            payload['name'] = name
        if description != None:
            payload['description'] = description
        if options != None:
            payload['options'] = options
        if default_permission != None:
            payload['default_permission'] = default_permission
        if type != None:
            payload['type'] = type
        return self._request('PATCH', params=payload, uri=f'/applications/{application_id}/guilds/{guild_id}/commands/{command_id}')

    def delete_guild_application_command(self, application_id, guild_id, command_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#delete-guild-application-command'''
        return self._request('DELETE', uri=f'/applications/{application_id}/guilds/{guild_id}/commands/{command_id}')

    def bulk_overwrite_guild_application_command(self, application_id, guild_id, new_commands: list=None) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#bulk-overwrite-guild-application-commands'''
        params = {}
        if new_commands != None:
            params['new_commands'] = new_commands
        return self._request('PUT', params=params, uri=f'/applications/{application_id}/guilds/{guild_id}/commands')

    def get_guild_application_command_permissions(self, application_id, guild_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-guild-application-command-permissions'''
        return self._request('GET', uri=f'/applications/{application_id}/guilds/{guild_id}/commands/permissions')

    def get_application_comman_permissions(self, application_id, guild_id, command_id) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#get-application-command-permissions'''
        return self._request('GET', uri=f'/applications/{application_id}/guilds/{guild_id}/commands/{command_id}/permissions')

    def edit_application_command_permissions(self, application_id, guild_id, command_id, permissions: list) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#edit-application-command-permissions'''
        payload = { 'permissions': permissions }
        return self._request('PUT', uri=f'/applications/{application_id}/guilds/{guild_id}/commands/{command_id}/permissions')

    def batch_edit_application_command_permissions(self, application_id, guild_id, commands: [dict]) -> dict:
        '''https://discord.com/developers/docs/interactions/application-commands#batch-edit-application-command-permissions'''
        return self._request('PUT', params=commands, uri=f'/applications/{application_id}/guilds/{guild_id}/commands/permissions')







class Webhook(object):
    '''Contains a collection of webhook related methods.'''
    
    async def create_webhook(self, channel_id, name: str, avatar=None) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#create-webhook'''
        payload = { 'name': name }
        if avatar != None:
            payload['avatar'] = avatar
        return await self._request('POST', params=payload, uri=f'/channels/{channel_id}/webhooks')

    async def get_channel_webhooks(self, channel_id) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#create-webhook'''
        return await self._request('GET', uri=f'/channels/{channel_id}/webhooks')

    async def get_guild_webhooks(self, guild_id) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#get-guild-webhooks'''
        return await self._request('GET', uri=f'/guilds/{guild_id}/webhooks')

    async def get_webhook(self, webhook_id) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#get-webhook'''
        return await self._request('GET', uri=f'/webhooks/{webhook_id}')

    async def get_webhook_with_token(self, webhook_id, webhook_token: str) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#get-webhook-with-token'''
        return await self._request('GET', uri=f'/webhooks/{webhook_id}/{webhook_token}')

    async def modify_webhook(self, webhook_id, name: str=None, avatar=None, channel_id=None) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#modify-webhook'''
        payload = {}
        if name != None:
            payload['name'] = name
        if avatar != None:
            payload['avatar'] = avatar
        if channel_id != None:
            payload['channel_id'] = channel_id
        return await self._request('PATCH', params=payload, uri=f'/webhooks/{webhook_id}')

    async def modify_webhook_with_token(self, webhook_id, webhook_token, name: str=None, avatar=None) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#modify-webhook-with-token'''
        payload = {}
        if name != None:
            payload['name'] = name
        if avatar != None:
            payload['avatar'] = avatar
        return await self._request('PATCH', params=payload, uri=f'/webhooks/{webhook_id}/{webhook_token}')

    async def delete_webhook(self) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#delete-webhook'''
        return await self._request('DELETE', uri=f'/webhooks/{webhook_id}')

    async def delete_webhook_with_token(self, webhook_id, webhook_token) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#delete-webhook-with-token'''
        return await self._request('DELETE', uri=f'/webhooks/{webhook_id}/{webhook_token}')

    async def execute_webhook(self, 
        wait: bool=False, 
        thread_id=None, 
        content: str=None, 
        username: str=None,
        avatar_url: str=None,
        tts: bool=None,
        embeds: list=None,
        allowed_mentions=None,
        components: list=None,
        files=None,
        payload_json: dict=None,
        attachments: list=None,
        flags: int=None
    ) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#execute-webhook
        ! Note that when sending a message, you must provide a value for at least one of content, embeds, or file.
        '''
        payload = {}
        if content != None:
            payload['content'] = content
        if username != None:
            payload['username'] = username
        if avatar_url != None:
            payload['avatar_url'] = avatar_url
        if tts != None:
            payload['tts'] = tts
        if allowed_mentions != None:
            payload['allowed_mentions'] = allowed_mentions
        if components != None:
            payload['components'] = components
        if files != None:
            payload['files'] = files
        if payload_json != None:
            payload['payload_json'] = payload_json
        if attachments != None:
            payload['attachments'] = attachments
        if flags != None:
            payload['flags'] = flags

        uri = f'/webhooks/{webhook_id}/{webhook_token}'
        if wait or thread_id != None:
            uri = f'{uri}?{wait}&thread_id={thread_id}'
        return await self._request('POST', params=payload, uri=uri)

    async def execute_slack_compatible_webhook(self) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#execute-slackcompatible-webhook'''
        raise NotImplementedError()

    async def execute_github_compatible_webhook(self) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#execute-githubcompatible-webhook'''
        raise NotImplementedError()

    async def get_webhook_messages(self, webhook_id, webhook_token, message_id, thread_id=None) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#get-webhook-message'''
        params = {}
        if thread_id != None:
            params['thread_id'] = thread_id
        return await self._request('GET', params=params, uri=f'/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}')

    async def edit_webhook_messages(self) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#edit-webhook-message'''
        raise NotImplementedError()

    async def delete_webhook_messages(self, webhook_id, webhook_token, message_id) -> dict:
        '''https://discord.com/developers/docs/resources/webhook#delete-webhook-message'''
        return await self._request('DELETE', uri=f'/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}')

    async def get_original_interaction_response(self, application_id, interaction_token, thread_id=None) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#get-original-interaction-response'''
        params = {}
        if thread_id != None:
            params['thread_id'] = thread_id
        return await self._request('GET', params=params, uri=f'/webhooks/{application_id}/{interaction_token}/messages/@original')

    async def edit_original_interaction_response(self, application_id, interaction_token) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#edit-original-interaction-response'''
        raise NotImplementedError()
        #return await self._request('PATCH', uri=f'/webhooks/{application_id}/{interaction_token}/messages/@original')
        
    async def delete_original_interaction_response(self, application_id, interaction_token) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#delete-original-interaction-response'''
        return await self._request('DELETE', uri=f'/webhooks/{application_id}/{interaction_token}/messages/@original')

    async def create_followup_message(self, 
        application_id, 
        interaction_token,
        content: str=None, 
        tts: bool=None,
        embeds: list=None,
        allowed_mentions=None,
        components: list=None,
        files=None,
        payload_json: dict=None,
        attachments: list=None,
        flags: int=None
    ) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#create-followup-message'''
        payload = {}
        if content != None:
            payload['content'] = content
        if tts != None:
            payload['tts'] = tts
        if allowed_mentions != None:
            payload['allowed_mentions'] = allowed_mentions
        if components != None:
            payload['components'] = components
        if files != None:
            payload['files'] = files
        if payload_json != None:
            payload['payload_json'] = payload_json
        if attachments != None:
            payload['attachments'] = attachments
        if flags != None:
            payload['flags'] = flags

        return await self._request('POST', params=payload, uri=f'/webhooks/{application_id}/{interaction_token}')

    async def get_followup_message(self, application_id, interaction_token, message_id, thread_id=None) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#get-followup-message'''
        params = {}
        if thread_id != None:
            params['thread_id'] = thread_id
        return await self._request('GET', params=params, uri=f'/webhooks/{application_id}/{interaction_token}/messages/{message_id}')

    async def edit_followup_message(self, application_id, interaction_token, message_id) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#edit-followup-message'''
        raise NotImplementedError()
        #return await self._request('PATCH', uri=f'/webhooks/{application_id}/{interaction_token}/messages/{message_id}')

    async def delete_followup_message(self, application_id, interaction_token, message_id) -> dict:
        '''https://discord.com/developers/docs/interactions/receiving-and-responding#delete-followup-message'''
        return await self._request('DELETE', uri=f'/webhooks/{application_id}/{interaction_token}/messages/{message_id}')

